#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "const.h"
#include "nlllib.h"
#include "formula.h"
#include "command.h"
#include "line.h"
#include "label.h"
#include "value.h"
#include "string.h"
#include "variable.h"
#include "position.h"
#include "array.h"
#include "area.h"
#include "stack.h"
#include "memory.h"
#include "function.h"
#include "readline.h"
#include "key.h"
#include "framebuf.h"
#include "mouse.h"
#include "graphic.h"
#include "image.h"
#include "nll.h"

static int stepping = 0;
static int skipping = 0;

static struct spot break_spot;
static struct spot continue_spot;

static int clear_break(line_t line)
{
  if (!line || (break_spot.line == line)) {
    break_spot.line = NULL;
    break_spot.command = NULL;
  }
  return 0;
}

static int clear_continue(line_t line)
{
  if (!line || (continue_spot.line == line)) {
    continue_spot.line = NULL;
    continue_spot.command = NULL;
  }
  return 0;
}

int nll_clear_line(line_t line)
{
  clear_break(line);
  clear_continue(line);
  return 0;
}

static int get_continue(spot_t spot)
{
  if (continue_spot.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;
  spot->line    = continue_spot.line;
  spot->command = continue_spot.command;
  return 0;
}

static int set_continue(spot_t spot)
{
  continue_spot.line    = spot->line;
  continue_spot.command = spot->command;
  return 0;
}

int nll_init(FILE *in, FILE *out)
{
  nll_stdin  = in;
  nll_stdout = out;

  nll_clear_line(NULL);

  memory_init();
  variable_init();
  value_init();
  string_init();
  position_init();
  label_init();
  formula_init();
  command_init();
  line_init();
  array_init();
  area_init();
  stack_init();
  framebuf_init();
#ifdef USE_GRAPHIC
  image_init();
  graphic_init();
#endif
  return 0;
}

int nll_done(void)
{
  int r;

  framebuf_done();
#ifdef USE_GRAPHIC
  graphic_done();
  image_done();
#endif

  stack_clear();
  line_clear();
  variable_clear();

#ifdef USE_GRAPHIC
  if ((r = graphic_check() ) < 0) return r;
  if ((r = image_check()   ) < 0) return r;
#endif
  if ((r = memory_check()  ) < 0) return r;
  if ((r = variable_check()) < 0) return r;
  if ((r = value_check()   ) < 0) return r;
  if ((r = string_check()  ) < 0) return r;
  if ((r = label_check()   ) < 0) return r;
  if ((r = formula_check() ) < 0) return r;
  if ((r = array_check()   ) < 0) return r;
  if ((r = command_check() ) < 0) return r;
  if ((r = line_check()    ) < 0) return r;
  if ((r = area_check()    ) < 0) return r;
  if ((r = stack_check()   ) < 0) return r;

  nll_file_close();
  nll_stdin_restore();
  nll_stdout_restore();

  nll_stdin  = NULL;
  nll_stdout = NULL;

  return 0;
}

static int _get_word(char **pp, const char *separator, char *buffer, int size)
{
  char c, *p;

  while (isspace(**pp))
    (*pp)++;

  p = buffer;

  while (1) {
    c = **pp;
    if (c == '\0') {
      break;
    } else if (strchr(separator, c)) {
      break;
    }
    **pp = toupper(c);
    if (p + 1 >= buffer + size)
      return NLL_ERRCODE_NLL_WORD_TOO_LONG;
    *(p++) = **pp;
    (*pp)++;
  }

  *p = '\0';

  return 0;
}

static int get_word(char **pp, char *buffer, int size)
{
  return _get_word(pp, ", :;(\t\n\r", buffer, size);
}

static int proc_end(command_arg_t args, int *retcodep)
{
  int r;
  integer_t retcode = 0;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &retcode)) < 0)
	return r;
    }
  }

  if (retcodep)
    *retcodep = retcode;

  return 0;
}

static int proc_args(command_arg_t args)
{
  int r, i;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, NULL)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_let(command_arg_t args)
{
  int r, i;
  value_t value = NULL, v;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &v)) < 0)
      return r;
    if ((r = value_copy_values(value, v)) < 0)
      ;
  }

  for (i = 2; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, NULL)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_print(command_arg_t args)
{
  int r, i, printed = 0;
  value_t value;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      if (value) {
	value_output_values(value, " ");
	framebuf_puts("\n");
	printed = 1;
      }
    }
  }

  if (!printed)
    framebuf_puts("\n");

  return 0;
}

static int proc_goto(spot_t spot, command_arg_t args)
{
  int r;
  value_t value, v;

  if ((args[0].type != COMMAND_ARG_LABEL) ||
      !args[0].arg.label.label->spot.line)
    return NLL_ERRCODE_LABEL_NOT_FOUND;

  spot->line    = args[0].arg.label.label->spot.line;
  spot->command = args[0].arg.label.label->spot.command;

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
      return r;
  }

  if ((r = formula_proc(spot->command->args[0].arg.label.args, &value)) < 0)
    return r;

  if ((r = formula_proc(args[0].arg.label.args, &v)) < 0)
    return r;
  if ((r = value_copy_values(value, v)) < 0)
    return r;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &v)) < 0)
      return r;
    if ((r = value_copy_values(value, v)) < 0)
      return r;
  }

  return 0;
}

static int proc_gosub(spot_t spot, command_arg_t args)
{
  int r;
  value_t value, v;

  if ((args[0].type != COMMAND_ARG_LABEL) ||
      !args[0].arg.label.label->spot.line)
    return NLL_ERRCODE_LABEL_NOT_FOUND;

  spot->line    = args[0].arg.label.label->spot.line;
  spot->command = args[0].arg.label.label->spot.command;

  if ((r = formula_proc(spot->command->args[0].arg.label.args, &value)) < 0)
    return r;

  if ((r = formula_proc(args[0].arg.label.args, &v)) < 0)
    return r;
  if ((r = value_copy_values(value, v)) < 0)
    return r;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &v)) < 0)
      return r;
    if ((r = value_copy_values(value, v)) < 0)
      return r;
  }

  return 0;
}

static int proc_gonext(spot_t spot, spot_t current, line_t line, command_arg_t args)
{
  int r;
  integer_t n = 1, n2 = -1, search = 0;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n)) < 0)
	return r;
      search = 1;
    }
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n2)) < 0)
	return r;
    }
  }

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (n2 < 0) {
    n2 = 0;
    if (!search)
      search = 1;
  }

  spot->line    = current->line;
  spot->command = current->command;

  if (search) {
    if (n < 0) {
      for (; spot->line && (n < 0); n++)
	spot->line = spot->line->prev;
    } else {
      for (; spot->line && (n > 0); n--)
	spot->line = spot->line->next;
    }
    spot->command = spot->line ? spot->line->commands : NULL;
  }

  while (n2 > 0) {
    if (!spot->line)
      break;
    if (!spot->command) {
      spot->line = spot->line->next;
      spot->command = spot->line ? spot->line->commands : NULL;
      continue;
    }
    spot->command = spot->command->next;
    n2--;
  }

  return 0;
}

static int proc_return(spot_t spot, command_arg_t args)
{
  int r;
  value_t value = NULL, v;
  position_type_t type;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
      return r;
  }

  type = position_get_type();

  switch (type) {
  case POSITION_TYPE_GOSUB:
    if ((r = position_pop(POSITION_TYPE_GOSUB, spot, NULL)) < 0)
      return r;
    if (spot->line == NULL)
      return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;
    if (spot->command->args[2].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(spot->command->args[2].arg.formula.element, &v)) < 0)
	return r;
      if (args[0].type == COMMAND_ARG_FORMULA) {
	if ((r = value_copy_values(v, value)) < 0)
	  return r;
      }
    }
    return 1;

  case POSITION_TYPE_LABEL:
    if ((r = position_pop(POSITION_TYPE_LABEL, NULL, NULL)) < 0)
      return r;
    if ((r = position_push(POSITION_TYPE_LABEL, spot, value)) < 0)
      return r;
    return 0;

  case POSITION_TYPE_EVAL:
    if ((r = position_pop(POSITION_TYPE_EVAL, NULL, NULL)) < 0)
      return r;
    if ((r = position_push(POSITION_TYPE_EVAL, spot, value)) < 0)
      return r;
    return 0;

  default:
    break;
  }

  return NLL_ERRCODE_POSITION_DIFFERENT_TYPE;
}

static int proc_if(command_arg_t args)
{
  int r = 0;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if ((r = value_is_nzero(value)) < 0)
      return r;
  }

  if (r) {
    if (args[1].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
	return r;
    }
  } else {
    if (args[2].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_loop(spot_t spot, command_arg_t args)
{
  int r;
  value_t value, v;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, &v)) < 0)
      return r;
  }

  if (args[3].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[3].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (args[2].type == COMMAND_ARG_FORMULA) {
      if ((r = value_copy_values(value, v)) < 0)
	return r;
    } else {
      if (value) {
	if ((r = value_set_integer(value, 0)) < 0)
	  return r;
      }
    }
  }

  if ((r = position_push(POSITION_TYPE_LOOP, spot, NULL)) < 0)
    return r;

  return 0;
}

static int proc_for(spot_t spot, command_arg_t args)
{
  int r;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  if ((r = position_push(POSITION_TYPE_FOR, spot, NULL)) < 0)
    return r;

  return 0;
}

static int proc_loopend(spot_t spot, command_arg_t args)
{
  int r;
  integer_t integer = 0, count = -1, start = 0, step = 1, end = 1;
  struct spot s;
  value_t value = NULL;

  if ((r = position_pop(POSITION_TYPE_LOOP, &s, NULL)) < 0)
    return r;

  if (s.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, NULL)) < 0)
      return r;
  }

  args = s.command->args;

  if (value) {
    if ((args[0].type != COMMAND_ARG_FORMULA) ||
	!value_is_same(value, args[0].arg.formula.element->value))
      return NLL_ERRCODE_POSITION_DIFFERENT_VARIABLE;
  }

  value = NULL;
  if (args[0].type == COMMAND_ARG_FORMULA)
    value = args[0].arg.formula.element->value;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[1].arg.formula.element->value, &count)) < 0)
      return r;
  }
  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[2].arg.formula.element->value, &start)) < 0)
      return r;
  }
  if (args[3].type == COMMAND_ARG_FORMULA) {
    if ((r = value_get_integer(args[3].arg.formula.element->value, &step)) < 0)
      return r;
  }

  if (value) {
    if ((r = value_get_integer(value, &integer)) < 0)
      return r;

    integer += step;

    if ((r = value_set_integer(value, integer)) < 0)
      return r;

    end = start + step * count;
  }

  if ((count < 0) || (integer != end)) {
    if ((r = position_push(POSITION_TYPE_LOOP, &s, NULL)) < 0)
      return r;
    spot->line    = s.line;
    spot->command = s.command;
    return 1;
  }

  return 0;
}

static int proc_next(spot_t spot, command_arg_t args)
{
  int r;
  struct spot s;
  value_t value = NULL;

  if ((r = position_pop(POSITION_TYPE_FOR, &s, NULL)) < 0)
    return r;

  if (s.line == NULL)
    return NLL_ERRCODE_NLL_NO_CONTINUE_POINT;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, NULL)) < 0)
      return r;
  }

  args = s.command->args;

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, NULL)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
  }

  if (value) {
    if ((r = value_is_nzero(value)) < 0)
      return r;
    if (r)
      value = NULL;
  }

  if (!value) {
    if ((r = position_push(POSITION_TYPE_FOR, &s, NULL)) < 0)
      return r;
    spot->line    = s.line;
    spot->command = s.command;
    return 1;
  }

  return 0;
}

static int proc_discard(command_arg_t args)
{
  int r;
  integer_t n = 1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &n)) < 0)
	return r;
    }
  }

  for (; n > 0; n--) {
    if ((r = position_pop(POSITION_TYPE_NONE, NULL, NULL)) < 0)
      return r;
  }

  return 0;
}

static int make_array(value_t value, int *n)
{
  int r, i;
  array_t array;

  if (*n == 0)
    return 0;
  if ((r = array_get(&array, *n)) < 0)
    return r;
  for (i = 0; i < *n; i++) {
    if ((r = make_array(array[i].value, n + 1)) < 0)
      return r;
  }
  if ((r = value_set_array(value, *n, array)) < 0)
    return r;

  return 0;
}

static int proc_dim(command_arg_t args)
{
#define ARRAY_DIM_MAXNUM (COMMAND_ARG_MAXNUM - 1)
  int r, i, n[ARRAY_DIM_MAXNUM + 1];
  integer_t integer;
  value_t value;

  for (i = 0; i < ARRAY_DIM_MAXNUM; i++) {
    n[i] = 0;
    if (args[i + 1].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i + 1].arg.formula.element, &value)) < 0)
	return r;
      if (value) {
	if ((r = value_get_integer(value, &integer)) < 0)
	  return r;
	n[i] = integer;
      } else {
	n[i] = 1;
      }
    }
  }
  n[i] = 0;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = make_array(value, n)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_memory(command_arg_t args)
{
  int r;
  integer_t integer;
  value_t value, v;
  area_t area;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &v)) < 0)
      return r;
    if (args[0].type == COMMAND_ARG_FORMULA) {
      if ((r = value_get_integer(v, &integer)) < 0)
	return r;
      if ((r = area_alloc(&area, integer)) < 0)
	return r;
      if ((r = value_set_area(value, area)) < 0)
	return r;
    }
  }

  return 0;
}

static int proc_push(command_arg_t args)
{
  int r, i;
  value_t value;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      for (; value; value = value->next) {
	if ((r = stack_push(value)) < 0)
	  return r;
      }
    }
  }

  return 0;
}

static int proc_pop(command_arg_t args)
{
  int r, i, n = 0;
  value_t value;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      for (; value; value = value->next) {
	if ((r = stack_pop(value)) < 0)
	  return r;
      }
      n++;
    }
  }

  if (!n) {
    if ((r = stack_pop(NULL)) < 0)
      return r;
  }

  return 0;
}

static int proc_run(spot_t spot, command_arg_t args)
{
  int r;
  integer_t integer = 1;
  value_t value;
  line_t line = NULL;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  if ((r = line_get(integer, &line)) < 0)
    return r;

  stack_clean();
  line_clean();
  variable_clear();
  position_init();
  clear_continue(NULL);
  stepping = 0;
  skipping = 0;

  if (line) {
    spot->line    = line;
    spot->command = line->commands;
    return 1;
  }

  return 0;
}

static int proc_break(command_arg_t args)
{
  int r;
  integer_t integer0 = 0, integer1 = 0;
  value_t value;
  line_t line = NULL;
  command_t command;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer0)) < 0)
	return r;
    }
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer1)) < 0)
	return r;
    }
  }

  if ((r = line_get(integer0, &line)) < 0)
    return r;

  if (!line) {
    clear_break(NULL);
  } else {
    for (command = line->commands; command && (integer1 > 0); integer1--) {
      command = command->next;
    }
    break_spot.line    = line;
    break_spot.command = command;
  }

  return 0;
}

static int proc_step(command_arg_t args)
{
  int r;
  integer_t integer = 1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  stepping = integer;

  return 0;
}

static int proc_skip(command_arg_t args)
{
  int r;
  integer_t integer = 1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  skipping = integer;

  return 0;
}

static int proc_new(spot_t spot, command_arg_t args)
{
  int r, i, retval = 1;
  value_t value;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      if ((r = value_clear_value(value)) < 0)
	return r;
      retval = 0;
    }
  }

  return retval;
}

static int proc_list(command_arg_t args)
{
  int r;
  integer_t integer0 = -1, integer1 = -1;
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer0)) < 0)
	return r;
    }
  }

  integer1 = integer0;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer1)) < 0)
	return r;
    }
  }

  line_print(nll_stdout, integer0, integer1);

  return 0;
}

static int proc_save(command_arg_t args)
{
  int r, length;
  char *string = "sample.hob";
  integer_t integer0 = -1, integer1 = -1;
  value_t value;
  FILE *fp;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_string(value, &string, &length)) < 0)
	return r;
    }
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer0)) < 0)
	return r;
    }
  }

  integer1 = integer0;

  if (args[2].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[2].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer1)) < 0)
	return r;
    }
  }

  fp = fopen(string, "w");
  if (!fp)
    return NLL_ERRCODE_FILE_NOT_FOUND;

  line_print(fp, integer0, integer1);

  fclose(fp);

  return 0;
}

static int proc_load(command_arg_t args)
{
  int r, length;
  char *string = "sample.hob";
  value_t value;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_string(value, &string, &length)) < 0)
	return r;
    }
  }

  if ((r = nll_file_open(string)) < 0)
    return r;

  return 0;
}

static int proc_stdin(command_arg_t args)
{
  int r;
  char *filename = NULL;
  value_t value;
  FILE *fp;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_string(value, &filename, NULL)) < 0)
	return r;
    }
  }

  if (!filename) {
    if ((r = nll_stdin_restore()) < 0)
      return r;
  } else {
    fp = fopen(filename, "r");
    if (!fp)
      return NLL_ERRCODE_FILE_NOT_FOUND;
    if ((r = nll_stdin_save(fp)) < 0)
      return r;
  }

  return 0;
}

static int proc_stdout(command_arg_t args)
{
  int r;
  char *filename = NULL;
  value_t value;
  FILE *fp;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_string(value, &filename, NULL)) < 0)
	return r;
    }
  }

  if (!filename) {
    if ((r = nll_stdout_restore()) < 0)
      return r;
  } else {
    fp = fopen(filename, "w");
    if (!fp)
      return NLL_ERRCODE_FILE_NOT_FOUND;
    if ((r = nll_stdout_save(fp)) < 0)
      return r;
  }

  return 0;
}

static int proc_eval(spot_t spot, command_arg_t args)
{
  int r;
  char *string = NULL;
  value_t value, v;
  line_t line;
  struct spot s;

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &v)) < 0)
      return r;
    if (v) {
      if ((r = value_get_string(v, &string, NULL)) < 0)
	return r;
    }
  }

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
  }

  if (!string)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((r = line_set(-1, -1, string, &line)) < 0)
    return r;
  if (!line)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  if ((r = nll_parse(line)) < 0)
    goto err;

  s.line = line;
  s.command = line->commands;
  if ((r = nll_exec(&s, spot, POSITION_TYPE_EVAL, &v, NULL)) < 0)
    goto err;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = value_copy_values(value, v)) < 0)
      goto err;
  }

  if ((r = line_free(line)) < 0)
    return r;

  return 0;

err:
  if (line)
    line_free(line);
  return r;
}

static int proc_edit(command_arg_t args)
{
  int r, i;
  integer_t integer;
  value_t value;
  line_t line;

  for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
    if (args[i].type == COMMAND_ARG_FORMULA) {
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      if (value) {
	if ((r = value_get_integer(value, &integer)) < 0)
	  return r;

	if ((r = line_get(integer, &line)) < 0)
	  return r;

	if (!line)
	  line = line_tail();

	if (line) {
	  nll_edit_line(integer, line);
	}
      }
    }
  }

  return 0;
}

static int proc_wait(command_arg_t args)
{
  int r;
  integer_t integer = 100, integer2 = 0;
  value_t value;

  if (args[1].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[1].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer2)) < 0)
	return r;
      integer = 0;
    }
  }

  if (args[0].type == COMMAND_ARG_FORMULA) {
    if ((r = formula_proc(args[0].arg.formula.element, &value)) < 0)
      return r;
    if (value) {
      if ((r = value_get_integer(value, &integer)) < 0)
	return r;
    }
  }

  usleep(integer * 10000 + integer2 * 100);

  return 0;
}

static int proc_dump(command_arg_t args)
{
  int r, i, all = 1;
  integer_t flags;
  value_t value;

  for (i = 0; i < ARRAY_DIM_MAXNUM + 1; i++) {
    if (i == ARRAY_DIM_MAXNUM) {
      if (!all)
	break;
      flags = 0;
    } else {
      if (args[i].type != COMMAND_ARG_FORMULA)
	continue;
      if ((r = formula_proc(args[i].arg.formula.element, &value)) < 0)
	return r;
      if (!value)
	continue;
      if ((r = value_get_integer(value, &flags)) < 0)
	return r;
    }

    if (!flags)
      flags = NLL_D_ALL;

    if (flags & NLL_D_VARIABLE) variable_dump(nll_stdout, NULL);
    if (flags & NLL_D_STRING  ) string_dump(nll_stdout, NULL);
    if (flags & NLL_D_LABEL   ) label_dump(nll_stdout, NULL);
    if (flags & NLL_D_LINE    ) line_dump(nll_stdout, NULL);
    if (flags & NLL_D_STACK   ) stack_dump(nll_stdout);
    if (flags & NLL_L_COMMAND ) command_list(nll_stdout);
    if (flags & NLL_L_FUNCTION) function_list(nll_stdout);
    if (flags & NLL_L_CONST   ) variable_const_list(nll_stdout);
    all = 0;
  }

  return 0;
}

static void set_errcode(command_t command, int r)
{
  if (command)
    command->errcode = r;
}

#define ERRBREAK(r, command, param) \
	do { NLL_ERRPRINT((r), (param)); set_errcode((command), (r)); goto brk; } while (0)

int nll_parse(line_t line)
{
  char *p, *tail, *buffer, *pc;
#define ARG_LEN 256
  char arg0[ARG_LEN];
  int r, i, count = 0;
  command_t command = NULL, *tailp;
  struct spot s;

  buffer = line->buffer;
  p = buffer;
  tail = p + strlen(p);

  tailp = &line->commands;

  while (p < tail) {
    if (isspace(*p))
      goto cont;

    pc = p;

    if ((r = command_alloc(&command)) < 0) ERRBREAK(r, command, pc);
    command->line = pc;

    if (*p == '.') {
      p++;
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0)
	ERRBREAK(r, command, pc);
      if (arg0[0] == '\0')
	ERRBREAK(NLL_ERRCODE_LABEL_INVALID_NAME, command, pc);

      if ((r = command_get(COMMAND_LABEL, &command->ope)) < 0)
	ERRBREAK(r, command, pc);
      s.line    = line;
      s.command = command;
      if ((r = label_set(arg0, &s, &command->args[0].arg.label.label)) < 0)
	ERRBREAK(r, command, pc);
      command->args[0].type = COMMAND_ARG_LABEL;

      while (isspace(*p))
	p++;
      if (*p == ':') {
	p++;
	if ((r = formula_parse(p, &p, &command->args[0].arg.label.args, ";:")) < 0)
	  ERRBREAK(r, command, pc);
      }

      goto cont;
    }

    if ((r = command_search(p, &p, &command->ope)) < 0) {
      if ((r = formula_parse(p, &p, &command->prefix, NULL)) < 0)
	ERRBREAK(r, command, pc);
      if ((r = command_get(COMMAND_NONE, &command->ope)) < 0)
	ERRBREAK(r, command, pc);
      if (*p == ',') {
	p++;
	while (isspace(*p))
	  p++;
	if ((r = command_search(p, &p, &command->ope)) < 0) {
	  if ((r = command_get(COMMAND_ARGS, &command->ope)) < 0)
	    ERRBREAK(r, command, pc);
	}
      }
    }

    switch (command->ope->type) {
    case COMMAND_NONE:
      break;

    case COMMAND_COMMENT:
      goto brk;

    case COMMAND_STOP:
    case COMMAND_CONTINUE:
      break;

    case COMMAND_GOTO:
    case COMMAND_GOSUB:
      if ((r = get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, command, pc);
      if (arg0[0]) {
	if ((r = label_set(arg0, NULL, &command->args[0].arg.label.label)) < 0) ERRBREAK(r, command, pc);
	command->args[0].type = COMMAND_ARG_LABEL;
	while (isspace(*p))
	  p++;
#ifdef NLL_LABEL_ARGUMENTS
	if (*p == ':') {
	  p++;
	  if ((r = formula_parse(p, &p, &command->args[0].arg.label.args, NULL)) < 0)
	    ERRBREAK(r, command, pc);
	}
#endif
      }
      if (*p == ',') p++; else break;
      for (i = 1; i < COMMAND_ARG_MAXNUM; i++) {
	if ((r = formula_parse(p, &p, &command->args[i].arg.formula.element, NULL)) < 0)
	  ERRBREAK(r, command, pc);
	if (command->args[i].arg.formula.element)
	  command->args[i].type = COMMAND_ARG_FORMULA;
	if (*p == ',') p++; else break;
      }
      break;

    case COMMAND_END:
    case COMMAND_EXIT:
    case COMMAND_ARGS:
    case COMMAND_LET:
    case COMMAND_PRINT:
    case COMMAND_GONEXT:
    case COMMAND_RETURN:
    case COMMAND_IF:
    case COMMAND_LOOP:
    case COMMAND_LOOPEND:
    case COMMAND_FOR:
    case COMMAND_NEXT:
    case COMMAND_DISCARD:
    case COMMAND_DIM:
    case COMMAND_MEMORY:
    case COMMAND_PUSH:
    case COMMAND_POP:
    case COMMAND_RUN:
    case COMMAND_SKIP:
    case COMMAND_BREAK:
    case COMMAND_STEP:
    case COMMAND_NEW:
    case COMMAND_LIST:
    case COMMAND_SAVE:
    case COMMAND_LOAD:
    case COMMAND_STDIN:
    case COMMAND_STDOUT:
    case COMMAND_EVAL:
    case COMMAND_EDIT:
    case COMMAND_WAIT:
    case COMMAND_DUMP:
      for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
	if ((r = formula_parse(p, &p, &command->args[i].arg.formula.element, NULL)) < 0)
	  ERRBREAK(r, command, pc);
	if (command->args[i].arg.formula.element)
	  command->args[i].type = COMMAND_ARG_FORMULA;
	if (*p == ',') p++; else break;
      }
      break;

    case COMMAND_UNKNOWN:
    default:
      ERRBREAK(NLL_ERRCODE_COMMAND_NOT_FOUND, command, pc);
    }

  cont:
    if (command) {
      command->next = *tailp;
      *tailp = command;
      tailp = &(command->next);
      command = NULL;
      count++;
    }

    if (*p)
      p++;
  }

  r = count;

brk:
  if (command) {
    if (!command->errcode) {
      command_free(command);
    } else {
      command_clear(command);
      command->next = *tailp;
      *tailp = command;
      tailp = &(command->next);
    }
  }

  return r;
}

static int nll_proc(spot_t spot, int *retcodep)
{
  line_t line;
  command_t command;
  int r, count = 0;
  struct spot s;
  value_t value;

  line    = spot->line;
  command = spot->command;

  for (; command; command = command->next) {
    if (skipping && (line_is_exist(line) > 0)) {
      skipping--;
      continue;
    }

    if (nll_is_finished()) {
      fprintf(nll_stdout, "\nBreak at: %s\n", command->line);
      goto brk;
    }

    if (!nll_finish_is_locked()) {
      if (stepping && (line_is_exist(line) > 0)) {
	stepping--;
	if (!stepping) {
	  fprintf(nll_stdout, "\nStep break at: %s\n", command->line);
	  goto brk;
	}
      }
    }

    if (break_spot.line && (line_is_exist(line) > 0)) {
      if ((line == break_spot.line) && (command == break_spot.command)) {
	clear_break(line);
	fprintf(nll_stdout, "\nBreakpoint at: %s\n", command->line);
	goto brk;
      }
    }

    if (command->errcode)
      ERRBREAK(command->errcode, NULL, command->line);

    count++;

    if (command->prefix) {
      if ((r = formula_proc(command->prefix, &value)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (command->ope->type == COMMAND_NONE)
	continue;
      if (!value) {
	r = 0;
      } else {
	if ((r = value_is_nzero(value)) < 0)
	  ERRBREAK(r, NULL, command->line);
      }
      if (!r)
	continue;
    }

    switch (command->ope->type) {
    case COMMAND_END:
      if ((r = proc_end(command->args, retcodep)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto end;

    case COMMAND_EXIT:
      if ((r = proc_end(command->args, retcodep)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto ext;

    case COMMAND_ARGS:
      if ((r = proc_args(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_LET:
      if ((r = proc_let(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_PRINT:
      if ((r = proc_print(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_GOTO:
      if ((r = proc_goto(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_GONEXT:
      s.line    = line;
      s.command = command;
      if ((r = proc_gonext(spot, &s, line, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_GOSUB:
      if ((r = proc_gosub(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      s.line    = line;
      s.command = command;
      if ((r = position_push(POSITION_TYPE_GOSUB, &s, NULL)) < 0)
	ERRBREAK(r, NULL, command->line);
      goto next;

    case COMMAND_RETURN:
      if ((r = proc_return(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	goto end;
      spot->command = spot->command->next;
      goto next;

    case COMMAND_IF:
      if ((r = proc_if(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_LOOP:
      s.line    = line;
      s.command = command;
      if ((r = proc_loop(&s, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_FOR:
      s.line    = line;
      s.command = command;
      if ((r = proc_for(&s, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_LOOPEND:
      if ((r = proc_loopend(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      spot->command = spot->command->next;
      goto next;

    case COMMAND_NEXT:
      if ((r = proc_next(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      spot->command = spot->command->next;
      goto next;

    case COMMAND_DISCARD:
      if ((r = proc_discard(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_DIM:
      if ((r = proc_dim(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_MEMORY:
      if ((r = proc_memory(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_PUSH:
      if ((r = proc_push(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_POP:
      if ((r = proc_pop(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_RUN:
      if ((r = proc_run(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      goto next;

    case COMMAND_STOP:
      nll_finish();
      break;

    case COMMAND_BREAK:
      if ((r = proc_break(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_STEP:
      if ((r = proc_step(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (stepping) stepping++;
      goto cont;

    case COMMAND_CONTINUE:
      goto cont;

    case COMMAND_SKIP:
      if ((r = proc_skip(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_NEW:
      if ((r = proc_new(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      if (!r)
	break;
      goto init;

    case COMMAND_LIST:
      if ((r = proc_list(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_SAVE:
      if ((r = proc_save(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_LOAD:
      if ((r = proc_load(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_STDIN:
      if ((r = proc_stdin(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_STDOUT:
      if ((r = proc_stdout(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_EVAL:
      if ((r = proc_eval(spot, command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_EDIT:
      if ((r = proc_edit(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_WAIT:
      if ((r = proc_wait(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_DUMP:
      if ((r = proc_dump(command->args)) < 0)
	ERRBREAK(r, NULL, command->line);
      break;

    case COMMAND_NONE:
    case COMMAND_LABEL:
    case COMMAND_UNKNOWN:
    default:
      break;
    }
  }

  spot->line = line->next;
  if (spot->line)
    spot->command = spot->line->commands;

next:
  return 0;

cont:
  if ((r = get_continue(spot)) < 0)
    ERRBREAK(r, NULL, command->line);
  clear_continue(NULL);
  return 0;

end:
  spot->line = NULL;
  return 0;

init:
  spot->line = NULL;
  return 1;

brk:
  if (retcodep)
    *retcodep = 1;
  if (nll_finish_is_locked())
    goto ext;
  s.line    = line;
  s.command = command;
  set_continue(&s);
  spot->line = NULL;
  return 0;

ext:
  spot->line = NULL;
  return -1;
}

static int nll_run(spot_t spot, int *retcodep)
{
  int r = 0;

  if (retcodep)
    *retcodep = 0;

  while (spot->line) {
    r = nll_proc(spot, retcodep);
    fflush(nll_stdout);
  }

  return r;
}

int nll_start(line_t line, command_t command, int *retcodep)
{
  struct spot spot;
  int r;

  nll_finished_clear();

  spot.line = line;

  if (!line)
    spot.command = NULL;
  else
    spot.command = command ? command : line->commands;

  framebuf_enable(1);

  r = nll_run(&spot, retcodep);

  framebuf_enable(0);

  return r;
}

int nll_exec(spot_t spot, spot_t current, position_type_t type, value_t *valuep, int *retcodep)
{
  int r, errcode = 0, locked;

  if ((r = position_push(type, current, NULL)) < 0)
    return r;

  locked = nll_finish_lock(1);
  if (nll_run(spot, retcodep))
    errcode = NLL_ERRCODE_NLL_INVALID_COMMAND;
  nll_finish_lock(locked);

  if ((r = position_pop(type, NULL, valuep)) < 0)
    return r;

  return errcode;
}

char *nll_line_buffer(line_t line)
{
  return line->buffer;
}

int nll_line_edit(int number, int number2, char *buffer, line_t *linep)
{
  return line_set(number, number2, buffer, linep);
}

int nll_line_free(line_t line)
{
  return line_free(line);
}
